package ssw.mj.test;

import java.io.StringReader;

import ssw.mj.Scanner;
import ssw.mj.Token;

/**
 * Test cases for the <code>Scanner</code> class.
 * 
 * @author AW
 */
public class ScannerTest extends IdentifiedTestCase {
	private static final String CR = "\r";
	private static final String LF = "\n";
	
	private static TestPrintWriter output;
	
	// ---------- helpers ----------
	
	private void initScanner (String s) {
		output = new TestPrintWriter();
		Scanner.init(new StringReader(s), output);
	}
	
	private void checkNext (int kind, int line, int col) {
		checkNext(kind, line, col, 0, null);
	}
	
	private void checkNext (int kind, int line, int col, String str) {
		checkNext(kind, line, col, 0, str);
	}
	
	private void checkNext (int kind, int line, int col, int val, String str) {
		Token expected = new Token(kind, line, col, val, str);
		Token actual = Scanner.next();
		assertEquals(expected, actual);
	}

	// ---------------------------------------
	// ---------- test valid inputs ----------
	
	/** Tests all token types. */
	public void testAllTokens () {
		initScanner("anIdentifier 123 'c'" + LF +
		            "+ - * / % == != > >= < <= && || ++ -- ; , . ( ) [ ] { }" + LF +
		            "break class do else final if new print program read return void while" + LF);
		
		checkNext(Token.ident, 1, 1, "anIdentifier");
		checkNext(Token.number, 1, 14, 123, "123");
		checkNext(Token.charConst, 1, 18, 99, null);
		checkNext(Token.plus, 2, 1);
		checkNext(Token.minus, 2, 3);
		checkNext(Token.times, 2, 5);
		checkNext(Token.slash, 2, 7);
		checkNext(Token.rem, 2, 9);
		checkNext(Token.eql, 2, 11);
		checkNext(Token.neq, 2, 14);
		checkNext(Token.gtr, 2, 17);
		checkNext(Token.geq, 2, 19);
		checkNext(Token.less, 2, 22);
		checkNext(Token.leq, 2, 24);
		checkNext(Token.and, 2, 27);
		checkNext(Token.or, 2, 30);
		checkNext(Token.pplus, 2, 33);
		checkNext(Token.mminus, 2, 36);
		checkNext(Token.semicolon, 2, 39);
		checkNext(Token.comma, 2, 41);
		checkNext(Token.period, 2, 43);
		checkNext(Token.lpar, 2, 45);
		checkNext(Token.rpar, 2, 47);
		checkNext(Token.lbrack, 2, 49);
		checkNext(Token.rbrack, 2, 51);
		checkNext(Token.lbrace, 2, 53);
		checkNext(Token.rbrace, 2, 55);
		checkNext(Token.break_, 3, 1);
		checkNext(Token.class_, 3, 7);
		checkNext(Token.do_, 3, 13);
		checkNext(Token.else_, 3, 16);
		checkNext(Token.final_, 3, 21);
		checkNext(Token.if_, 3, 27);
		checkNext(Token.new_, 3, 30);
		checkNext(Token.print, 3, 34);
		checkNext(Token.program, 3, 40);
		checkNext(Token.read, 3, 48);
		checkNext(Token.return_, 3, 53);
		checkNext(Token.void_, 3, 60);
		checkNext(Token.while_, 3, 65);
		checkNext(Token.eof, 4, 0);

		output.verify();
	}
	
	/** Tests a short program with CR and LF for line separation. */
	public void testCRLFLineSeparators () {
		initScanner("1" + CR + LF +
		            " 2" + CR + LF +
		            "  3 ");
		
		checkNext(Token.number, 1, 1, 1, "1");
		checkNext(Token.number, 2, 2, 2, "2");
		checkNext(Token.number, 3, 3, 3, "3");
		checkNext(Token.eof, 3, 4);
		
		output.verify();
	}

	/** Tests a short program with only LF for line separation. */
	public void testLFLineSeparators() {
		initScanner("1" + LF +
		            " 2" + LF +
		            "  3 ");
		
		checkNext(Token.number, 1, 1, 1, "1");
		checkNext(Token.number, 2, 2, 2, "2");
		checkNext(Token.number, 3, 3, 3, "3");
		checkNext(Token.eof, 3, 4);
		
		output.verify();
	}

	/** Tests a short program with one single line comment. */
	public void testSimpleSingleLineComment () {
		initScanner("class SSLComment { /* Simple single line comment. */ }  ");

		checkNext(Token.class_, 1, 1, "class");
		checkNext(Token.ident, 1, 7, "SSLComment");
		checkNext(Token.lbrace, 1, 18);
		checkNext(Token.rbrace, 1, 54);
		checkNext(Token.eof, 1, 56);
		
		output.verify();
	}
	
	/** Tests a short program with one multi-line comment. */
	public void testSimpleMultiLineComment () {
		initScanner("class SMLComment {" + LF +
		            "/*\tThis " + LF +
		            "\tis " + LF +
		            "\ta " + LF +
		            "\tsimple " + LF +
		            "\tmulti-line " + LF +
		            "\t comment. */ " + LF +
		            " }");

		checkNext(Token.class_, 1, 1, "class");
		checkNext(Token.ident, 1, 7, "SMLComment");
		checkNext(Token.lbrace, 1, 18);
		checkNext(Token.rbrace, 8, 2);
		checkNext(Token.eof, 8, 2);
		
		output.verify();
	}

	/** Tests a short program with nested single line comments. */
	public void testNestedSingleLineComment () {
		initScanner("class NSLComment {" + LF +
		            " /* This is a /* nested  /* single-line */ comment. */*/ " + LF +
		            "}");

		checkNext(Token.class_, 1, 1, "class");
		checkNext(Token.ident, 1, 7, "NSLComment");
		checkNext(Token.lbrace, 1, 18);
		checkNext(Token.rbrace, 3, 1);
		checkNext(Token.eof, 3, 1);		
		
		output.verify();
	}

	/** Tests a short program with nested multi-line comments. */
	public void testNestedMultiLineComment () {
		initScanner("class NMLComment {" + LF +
		            "/*" + LF +
		            "\tThis is a " + LF +
		            "\t/*\tnested  " + LF +
		            "\t\t/* multi-line */" + LF +
		            "\t\tcomment. " + LF +
		            "\t*/" + LF +
		            "*/ " + LF +
		            "}");

		checkNext(Token.class_, 1, 1, "class");
		checkNext(Token.ident, 1, 7, "NMLComment");
		checkNext(Token.lbrace, 1, 18);
		checkNext(Token.rbrace, 9, 1);
		checkNext(Token.eof, 9, 1);
		
		output.verify();
	}

  /** Tests some valid character constants. */
	public void testCharConstants () {
		initScanner("space ' ' 'A' 'z' '0' '!' '/' '@' '^' '|' '~' '\\'");
		
		checkNext(Token.ident, 1, 1, "space");
		checkNext(Token.charConst, 1, 7, 32, null);	
		checkNext(Token.charConst, 1, 11, 65, null);	
		checkNext(Token.charConst, 1, 15, 122, null);	
		checkNext(Token.charConst, 1, 19, 48, null);	
		checkNext(Token.charConst, 1, 23, 33, null);	
		checkNext(Token.charConst, 1, 27, 47, null);	
		checkNext(Token.charConst, 1, 31, 64, null);	
		checkNext(Token.charConst, 1, 35, 94, null);	
		checkNext(Token.charConst, 1, 39, 124, null);	
		checkNext(Token.charConst, 1, 43, 126, null);	
		checkNext(Token.charConst, 1, 47, 92, null);
		
		output.verify();
	}

  /** Tests all valid ESC character sequences. */	
	public void testESCsequences () {
		initScanner("backslash '\\'" + LF +
		            "linefeed '\\n'" + LF +
		            "carriage return '\\r'");
		
		checkNext(Token.ident, 1, 1, "backslash");
		checkNext(Token.charConst, 1, 11, 92, null);
		checkNext(Token.ident, 2, 1, "linefeed");
		checkNext(Token.charConst, 2, 10, 10, null);
		checkNext(Token.ident, 3, 1, "carriage");
		checkNext(Token.return_, 3, 10, "return");
		checkNext(Token.charConst, 3, 17, 13, null);
		checkNext(Token.eof, 3, 20);
		
		output.verify();
	}
	
	/** Tests invalid characters. */	
	public void testCharErrors () {
		initScanner("'A;" + LF +
		            "'\\A';" + LF +
		            "'\\A;" + LF +
		            "'\\;" + LF + 
		            "'\\\\';");
		
		checkNext(Token.charConst, 1, 1, 65, null);
		output.addExpectedLine("-- line 1, col 3: ", "MISS_SQUOTE");
		checkNext(Token.semicolon, 1, 3);
		
		checkNext(Token.charConst, 2, 1, 92, null);
		output.addExpectedLine("-- line 2, col 3: ", "UNDEF_ESC", " 'A'");
		checkNext(Token.semicolon, 2, 5);

		checkNext(Token.charConst, 3, 1, 92, null);
		output.addExpectedLine("-- line 3, col 3: ", "UNDEF_ESC", " 'A'");
		output.addExpectedLine("-- line 3, col 4: ", "MISS_SQUOTE");
		checkNext(Token.semicolon, 3, 4);
		
		checkNext(Token.charConst, 4, 1, 92, null);
		output.addExpectedLine("-- line 4, col 3: ", "UNDEF_ESC", " ';'");
		output.addExpectedLine("-- line 5, col 0: ", "MISS_SQUOTE");
		// ";" is treated as the escape character, so no semicolon-token expected

		checkNext(Token.charConst, 5, 1, 92, null);
		output.addExpectedLine("-- line 5, col 3: ", "UNDEF_ESC", " '\\'");
		checkNext(Token.semicolon, 5, 5);

		output.verify();
	}

	/** Tests number followed by identifier. */
    public void testNumberIdent() {
        initScanner("123abc;" + LF + 
                    "123while;");

        checkNext(Token.number, 1, 1, 123, "123");
        checkNext(Token.ident, 1, 4, 0, "abc");
        checkNext(Token.semicolon, 1, 7);
        
        checkNext(Token.number, 2, 1, 123, "123");
        checkNext(Token.while_, 2, 4);
        checkNext(Token.semicolon, 2, 9);

        output.verify();
    }

	// ----------------------------------------
	// ---------- test invalid input ----------	
	
	/** Tests some invalid characters. */
	public void testInvalidSymbols () {
		initScanner("ch = '';");
		
		checkNext(Token.ident, 1, 1, "ch");
		checkNext(Token.assign, 1, 4);
		checkNext(Token.charConst, 1, 6, 0, null);
		output.addExpectedLine("-- line 1, col 7: ", "INVALID_CHARCONST", " ''");
		checkNext(Token.semicolon, 1, 9);
		checkNext(Token.eof, 1, 9);
		
		output.verify();
		
		initScanner("''");
		
		checkNext(Token.charConst, 1, 1, 0, null);
		output.addExpectedLine("-- line 1, col 2: ", "INVALID_CHARCONST", " ''");
		checkNext(Token.eof, 1, 3);
		
		output.verify();
		
		initScanner("");
		
		checkNext(Token.none, 1, 1);
		output.addExpectedLine("-- line 1, col 1: ", "INVALID_CHAR", " ''");
		checkNext(Token.eof, 1, 1);
		
		output.verify();
	}

	/** Tests an invalid ESC character sequence. */
	public void testInvalidESCsequence () {
		initScanner("unknown '\\x'");
		
		checkNext(Token.ident, 1, 1, "unknown");
		checkNext(Token.charConst, 1, 9, 92, null);
		output.addExpectedLine("-- line 1, col 11: ", "UNDEF_ESC", " 'x'");
		checkNext(Token.eof, 1, 12);
		
		output.verify();
	}

	/** Tests character constant without closing single quote. */
	public void testNoClosingQuote () {
		initScanner("'x;");
		
		checkNext(Token.charConst, 1, 1, 'x', null);
		output.addExpectedLine("-- line 1, col 3: ", "MISS_SQUOTE");
		checkNext(Token.semicolon, 1, 3);
		checkNext(Token.eof, 1, 3);
		
		output.verify();
	}
	
	/** Tests empty character constant. */
	public void testEmptyCharConst () {
		initScanner("char ch = '';");
		
		checkNext(Token.ident, 1, 1, "char");
		checkNext(Token.ident, 1, 6, "ch");
		checkNext(Token.assign, 1, 9);
		checkNext(Token.charConst, 1, 11, 0, null);
		output.addExpectedLine("-- line 1, col 12: ", "EMPTY_CHARCONST");
		checkNext(Token.semicolon, 1, 13);
		checkNext(Token.eof, 1, 13);
		
		output.verify();
	}
	
	/** Tests large and too large integer constants. */
	public void testHugeNumbers () {
		initScanner("n = 2147483647;" + LF +
		            "n = 2147483648;" + LF +
		            "n = 2147483647 + 1;" + LF +
		            "n = -2147483648;" + LF +
								"n = -2147483647 - 1;");
		
		checkNext(Token.ident, 1, 1, "n");
		checkNext(Token.assign, 1, 3);
		checkNext(Token.number, 1, 5, 2147483647, "2147483647");
		checkNext(Token.semicolon, 1, 15);

		checkNext(Token.ident, 2, 1, "n");
		checkNext(Token.assign, 2, 3);
		checkNext(Token.number, 2, 5, 0, "2147483648");
		output.addExpectedLine("-- line 2, col 15: ", "BIG_NUM", ": 2147483648");
		checkNext(Token.semicolon, 2, 15);

		checkNext(Token.ident, 3, 1, "n");
		checkNext(Token.assign, 3, 3);
		checkNext(Token.number, 3, 5, 2147483647, "2147483647");
		checkNext(Token.plus, 3, 16);
		checkNext(Token.number, 3, 18, 1, "1");
		checkNext(Token.semicolon, 3, 19);

		checkNext(Token.ident, 4, 1, "n");
		checkNext(Token.assign, 4, 3);
		checkNext(Token.minus, 4, 5);
		checkNext(Token.number, 4, 6, 0, "2147483648");
		output.addExpectedLine("-- line 4, col 16: ", "BIG_NUM", ": 2147483648");
		checkNext(Token.semicolon, 4, 16);
		
		checkNext(Token.ident, 5, 1, "n");
		checkNext(Token.assign, 5, 3);
		checkNext(Token.minus, 5, 5);
		checkNext(Token.number, 5, 6, 2147483647, "2147483647");
		checkNext(Token.minus, 5, 17);
		checkNext(Token.number, 5, 19, 1, "1");
		checkNext(Token.semicolon, 5, 20);
		checkNext(Token.eof, 5, 20);
		
		output.verify();
	}

	// ----------------------------------------------------
	// ---------- test complete correct programs ----------

	/** Tests the shortest valid program: */
	public void testShortestProgram () {
		initScanner("class Shortest { }");

		checkNext(Token.class_, 1, 1, "class");
		checkNext(Token.ident, 1, 7, "Shortest");
		checkNext(Token.lbrace, 1, 16, null);
		checkNext(Token.rbrace, 1, 18, null);
		checkNext(Token.eof, 1, 18, null);
		
		output.verify();
	}

	/** Tests DoNothing program which already contains some basic constructs.
	 *  This test program uses tabulator characters for indentation.  
	 */
	public void testDoNothingProgram () {
		initScanner("class DoNothing" + LF +
		            "\tint dummy;" + LF +
		            "{" + LF +
		            "\tint doNothing() { dummy=0; return dummy; }" + LF +
		            "" + LF +
		            "\tvoid main() { " + LF +
		            "\t\tdummy= doNothing();" + LF +
		            "\t}" + LF +
		            "}");

		checkNext(Token.class_, 1, 1, "class");
		checkNext(Token.ident, 1, 7, "DoNothing");
		checkNext(Token.ident, 2, 2, "int");
		checkNext(Token.ident, 2, 6, "dummy");
		checkNext(Token.semicolon, 2, 11);
		checkNext(Token.lbrace, 3, 1);
		checkNext(Token.ident, 4, 2, "int");
		checkNext(Token.ident, 4, 6, "doNothing");
		checkNext(Token.lpar, 4, 15);
		checkNext(Token.rpar, 4, 16);
		checkNext(Token.lbrace, 4, 18);
		checkNext(Token.ident, 4, 20, "dummy");
		checkNext(Token.assign, 4, 25);
		checkNext(Token.number, 4, 26, "0");
		checkNext(Token.semicolon, 4, 27);
		checkNext(Token.return_, 4, 29, "return");
		checkNext(Token.ident, 4, 36, "dummy");
		checkNext(Token.semicolon, 4, 41);
		checkNext(Token.rbrace, 4, 43);
		checkNext(Token.void_, 6, 2, "void");
		checkNext(Token.ident, 6, 7, "main");
		checkNext(Token.lpar, 6, 11);
		checkNext(Token.rpar, 6, 12);
		checkNext(Token.lbrace, 6, 14);
		checkNext(Token.ident, 7, 3, "dummy");
		checkNext(Token.assign, 7, 8);
		checkNext(Token.ident, 7, 10, "doNothing");
		checkNext(Token.lpar, 7, 19);
		checkNext(Token.rpar, 7, 20);
		checkNext(Token.semicolon, 7, 21);
		checkNext(Token.rbrace, 8, 2);
		checkNext(Token.rbrace, 9, 1);
		checkNext(Token.eof, 9, 1);

		output.verify();
	}

	/** Tests the example program from the lecture script (p. 8, 5.1). 
	 * 	This test program uses space characters for indentation. 
	 */
	public void testScriptExample () {
		initScanner("class P" + LF +
		            "" + LF +
		            "  final int size = 10;" + LF +
		            "" + LF +
		/*line  5*/ "  class Table {" + LF +
		            "    int[] pos;" + LF +
		            "    int[] neg;" + LF +
		            "  }" + LF +
		            "" + LF +
		/*line 10*/ "  Table val;" + LF +
		            "{" + LF +
		            "  void main()" + LF +
		            "    int x, i;" + LF +
		            "  { /*---------- Initialize val ------------*/" + LF +
		/*line 15*/ "    val = new Table;" + LF +
		            "    val.pos = new int[size]; val.neg = new int[size];" + LF +
		            "    i = 0;" + LF +
		            "    while (i < size) {" + LF +
		            "      val.pos[i] = 0; val.neg[i] = 0;" + LF +
		/*line 20*/ "      i++;" + LF +
		            "    }" + LF +
		            "    /*---------- Read values ---------*/" + LF +
		            "    read(x);" + LF +
		            "    while (x > 0) {" + LF +
		/*line 25*/ "      if (0 <= x && x < size) {" + LF +
		            "        val.pos[x]++;" + LF +
		            "      } else if (-size < x && x < 0) {" + LF +
		            "        val.neg[-x]++;" + LF +
		            "      }" + LF +
		/*line 30*/ "      read(x);" + LF +
		            "    }" + LF +
		            "  }" + LF +
		            "}");
		
		checkNext(Token.class_,1,1,"class"); checkNext(Token.ident,1,7,"P");
		checkNext(Token.final_,3,3,"final"); checkNext(Token.ident,3,9,"int"); checkNext(Token.ident,3,13,"size"); checkNext(Token.assign,3,18); checkNext(Token.number,3,20,10,"10"); checkNext(Token.semicolon,3,22);
		checkNext(Token.class_,5,3,"class"); checkNext(Token.ident,5,9,"Table"); checkNext(Token.lbrace,5,15);
		checkNext(Token.ident,6,5,"int"); checkNext(Token.lbrack,6,8); checkNext(Token.rbrack,6,9); checkNext(Token.ident,6,11,"pos"); checkNext(Token.semicolon,6,14);
		checkNext(Token.ident,7,5,"int"); checkNext(Token.lbrack,7,8); checkNext(Token.rbrack,7,9); checkNext(Token.ident,7,11,"neg"); checkNext(Token.semicolon,7,14);
		checkNext(Token.rbrace,8,3);
		checkNext(Token.ident,10,3,"Table"); checkNext(Token.ident,10,9,"val"); checkNext(Token.semicolon,10,12); 
		checkNext(Token.lbrace,11,1); 
		checkNext(Token.void_,12,3,"void"); checkNext(Token.ident,12,8,"main"); checkNext(Token.lpar,12,12); checkNext(Token.rpar,12,13);
		checkNext(Token.ident,13,5,"int"); checkNext(Token.ident,13,9,"x"); checkNext(Token.comma,13,10); checkNext(Token.ident,13,12,"i"); checkNext(Token.semicolon,13,13);
		checkNext(Token.lbrace,14,3);
		checkNext(Token.ident,15,5,"val"); checkNext(Token.assign,15,9); checkNext(Token.new_,15,11,"new"); checkNext(Token.ident,15,15,"Table"); checkNext(Token.semicolon,15,20);
		checkNext(Token.ident,16,5,"val"); checkNext(Token.period,16,8); checkNext(Token.ident,16,9,"pos"); checkNext(Token.assign,16,13); checkNext(Token.new_,16,15, "new"); checkNext(Token.ident,16,19,"int"); checkNext(Token.lbrack,16,22); checkNext(Token.ident,16,23,"size"); checkNext(Token.rbrack,16,27); checkNext(Token.semicolon,16,28);
		checkNext(Token.ident,16,30,"val"); checkNext(Token.period,16,33); checkNext(Token.ident,16,34,"neg"); checkNext(Token.assign,16,38); checkNext(Token.new_,16,40, "new"); checkNext(Token.ident,16,44,"int"); checkNext(Token.lbrack,16,47); checkNext(Token.ident,16,48,"size"); checkNext(Token.rbrack,16,52); checkNext(Token.semicolon,16,53);
		checkNext(Token.ident,17,5,"i"); checkNext(Token.assign,17,7); checkNext(Token.number,17,9,0,"0"); checkNext(Token.semicolon,17,10);
		checkNext(Token.while_,18,5,"while"); checkNext(Token.lpar,18,11); checkNext(Token.ident,18,12,"i"); checkNext(Token.less,18,14); checkNext(Token.ident,18,16,"size"); checkNext(Token.rpar,18,20); checkNext(Token.lbrace,18,22);
		checkNext(Token.ident,19,7,"val"); checkNext(Token.period,19,10); checkNext(Token.ident,19,11,"pos"); checkNext(Token.lbrack,19,14); checkNext(Token.ident,19,15,"i"); checkNext(Token.rbrack,19,16); checkNext(Token.assign,19,18); checkNext(Token.number,19,20,0,"0"); checkNext(Token.semicolon,19,21);
		checkNext(Token.ident,19,23,"val"); checkNext(Token.period,19,26); checkNext(Token.ident,19,27,"neg"); checkNext(Token.lbrack,19,30); checkNext(Token.ident,19,31,"i"); checkNext(Token.rbrack,19,32); checkNext(Token.assign,19,34); checkNext(Token.number,19,36,0,"0"); checkNext(Token.semicolon,19,37);
		checkNext(Token.ident,20,7,"i"); checkNext(Token.pplus,20,8); checkNext(Token.semicolon,20,10);
		checkNext(Token.rbrace,21,5);
		checkNext(Token.read,23,5,"read"); checkNext(Token.lpar,23,9); checkNext(Token.ident,23,10,"x"); checkNext(Token.rpar,23,11); checkNext(Token.semicolon,23,12);
		checkNext(Token.while_,24,5,"while"); checkNext(Token.lpar,24,11); checkNext(Token.ident,24,12,"x"); checkNext(Token.gtr,24,14); checkNext(Token.number,24,16,0,"0"); checkNext(Token.rpar,24,17); checkNext(Token.lbrace,24,19);
		checkNext(Token.if_,25,7,"if"); checkNext(Token.lpar,25,10); checkNext(Token.number,25,11,0,"0"); checkNext(Token.leq,25,13); checkNext(Token.ident,25,16,"x"); checkNext(Token.and,25,18); checkNext(Token.ident,25,21,"x"); checkNext(Token.less,25,23); checkNext(Token.ident,25,25,"size"); checkNext(Token.rpar,25,29); checkNext(Token.lbrace,25,31);
		checkNext(Token.ident,26,9,"val"); checkNext(Token.period,26,12); checkNext(Token.ident,26,13,"pos"); checkNext(Token.lbrack,26,16); checkNext(Token.ident,26,17,"x"); checkNext(Token.rbrack,26,18); checkNext(Token.pplus,26,19); checkNext(Token.semicolon,26,21);
		checkNext(Token.rbrace,27,7); checkNext(Token.else_,27,9,"else"); checkNext(Token.if_,27,14,"if"); checkNext(Token.lpar,27,17); checkNext(Token.minus,27,18); checkNext(Token.ident,27,19,"size"); checkNext(Token.less,27,24); checkNext(Token.ident,27,26,"x"); checkNext(Token.and,27,28); checkNext(Token.ident,27,31,"x"); checkNext(Token.less,27,33); checkNext(Token.number,27,35,0,"0"); checkNext(Token.rpar,27,36); checkNext(Token.lbrace,27,38);
		checkNext(Token.ident,28,9,"val"); checkNext(Token.period,28,12); checkNext(Token.ident,28,13,"neg"); checkNext(Token.lbrack,28,16); checkNext(Token.minus,28,17); checkNext(Token.ident,28,18,"x"); checkNext(Token.rbrack,28,19); checkNext(Token.pplus,28,20); checkNext(Token.semicolon,28,22);
		checkNext(Token.rbrace,29,7);
		checkNext(Token.read,30,7,"read"); checkNext(Token.lpar,30,11); checkNext(Token.ident,30,12,"x"); checkNext(Token.rpar,30,13); checkNext(Token.semicolon,30,14);
		checkNext(Token.rbrace,31,5);
		checkNext(Token.rbrace,32,3);
		checkNext(Token.rbrace,33,1);
		checkNext(Token.eof,33,1);

		output.verify();
	}	

	/** Starts a text-mode test runner that runs all scanner tests. */	
	public static void main (String[] args) {
		junit.textui.TestRunner.run(ScannerTest.class);	
	}
}
